/***************************************************************************/

/*
 *	pit.c -- Motorola ColdFire PIT timer. Currently this type of
 *	         hardware timer only exists in the Motorola ColdFire
 *		 5270/5271 and 5282 CPUs.
 *
 *	Copyright (C) 1999-2004, Greg Ungerer (gerg@snapgear.com)
 *	Copyright (C) 2001-2004, SnapGear Inc. (www.snapgear.com)
 *
 */

/***************************************************************************/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/param.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/coldfire.h>
#include <asm/mcfpit.h>
#include <asm/mcfsim.h>

/***************************************************************************/

void coldfire_pit_tick(void)
{
	volatile struct mcfpit *tp;

	/* Reset the ColdFire timer */
	tp = (volatile struct mcfpit *) (MCFPIT_BASE1);
	tp->pcsr |= MCFPIT_PCSR_PIF;

	// trigger the watchdog (no harm, if not activated)
	MCF_WTM_WSR=0x5555;
	MCF_WTM_WSR=0xAAAA;
}

/***************************************************************************/
/*
 * will be called after power on from the pm coldfire module
*/
void coldfire_restore_pit_after_powerfail(void)
{
	volatile struct mcfpit *tp;

	MCF_INTC1_ICR(MCFINT_PIT1)= 1; /* priority 1 */

	//imrp = (volatile unsigned long *) (MCF_IPSBAR + MCFICM_INTC0 + MCFINTC_IMRH);
	MCF_INTC1_IMRH &= ~(1 << (MCFINT_PIT1 - 32));

	/* Set up PIT timer 1 as poll clock */
	tp = (volatile struct mcfpit *) (MCFPIT_BASE1);
	tp->pcsr = MCFPIT_PCSR_DISABLE;

	//--------------------------------------
	//28.3.3 : timeout[s]= pre * (prm+1) / fsys [1/s]
	//--------------------------------------
	// timeout=1/HZ
	// pre=64
	// fsys=MCF_CLK
	//--------------------------------------
	// 1/HZ = 64 x (pmr+1) / MCF_CLK
	//--------------------------------------
	// pmr = (CLK * HZ / 64)-1
	//--------------------------------------
	tp->pmr = (MCF_BUSCLK/(64*HZ)) -1;
	tp->pcsr = MCFPIT_PCSR_EN | MCFPIT_PCSR_PIE | MCFPIT_PCSR_OVW |
		MCFPIT_PCSR_RLD | MCFPIT_PCSR_CLK64;
}
/***************************************************************************/

void coldfire_pit_init(irqreturn_t (*handler)(int, void *, struct pt_regs *))
{

	request_irq(64+64+MCFINT_PIT1, handler, SA_INTERRUPT, "ColdFire Timer", NULL);

	coldfire_restore_pit_after_powerfail();
}


/***************************************************************************/

unsigned long coldfire_pit_offset(void)
{
	volatile struct mcfpit *tp;
	volatile unsigned long *ipr;
	unsigned long pmr, pcntr, offset;

	tp = (volatile struct mcfpit *) (MCFPIT_BASE1);
	ipr = (volatile unsigned long *) &(MCF_INTC1_IMRH);

	pmr = *(&tp->pmr);
	pcntr = *(&tp->pcntr);

	/*
	 * If we are still in the first half of the upcount and a
	 * timer interupt is pending, then add on a ticks worth of time.
	 */
	offset = ((pmr - pcntr) * (1000000 / HZ)) / pmr;
	if ((offset < (1000000 / HZ / 2)) && (*ipr & (1 << (MCFINT_PIT1 - 32))))
		offset += 1000000 / HZ;
	return offset;
}

/***************************************************************************/
